"""
Copyright (c) 2020-2022 Go2Market Insights, LLC d/b/a Analyzr
All rights reserved.
https://analyzr.ai

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import requests
import json
# import bz2
import gzip
import logging
import webbrowser

from datetime import datetime, timedelta


logging.basicConfig(level=logging.INFO)
log = logging.getLogger("ApiClient")

TOKEN_REFRESH = 10  # every 10 days refresh jwt token
MAX_ATTEMPTS = 3 # number of login attempts by the client if a 408 response is received

class SamlSsoAuthClient:
    """
    The SamlSsoAuthClient class manages low-level interactions with the Analyzr API
    using SAML SSO authentication.

    """

    def __init__(self, host=None, verbose=False):
        """
        Initialize API client object

        :param host: FQDN for the API including schema, e.g. https://acme.api.g2m.ai
        :param verbose: Set to true for verbose output
        :return:
        """
        if verbose: log.info("Starting: Init Client")
        self.host = host
        self.url = None
        self.client_id = None
        self._token = {
            'token': ''
        }
        self._token_time = None
        return

    ############################################################################
    #                           Basic API calls                                #
    ############################################################################

    def _get(self, uri):
        """
        Returns standard GET response for URI. Note that this
        function is protected, not private.

        :param uri:
        :return res:
        """
        self._check_token()
        r = requests.get(uri, params=self._token)
        return self._response(r)

    def _post(self, uri, json_obj, compressed=False):
        """
        Returns standard POST response for URI. Note that this
        function is protected, not private.

        :param uri:
        :param json_obj:
        :param compressed:
        :return res:
        """
        data = json.dumps(json_obj)
        self._check_token()
        if compressed is False:
            r = requests.post(
                uri,
                json=data,
                params=self._token,
            )
        else:
            # Deprecated /compressed endpoint
            r = requests.post(
                uri,
                json=data,
                params=self._token,
            )
        return self._response(r)

    def _response(self, r):
        """
        """
        res = {}
        res['status'] = r.status_code
        if r.status_code>=200 and r.status_code<300:
            res['response'] = r.json()
        else:
            log.error('WARNING! Request returned status code: {}'.format(r.status_code))
            res['response'] = None
        return res


    ############################################################################
    #                      Authentication methods                              #
    ############################################################################

    def _login(self, attempts=0, verbose=False):
        """
        The client calls the API at <host> and handles the SAML SSO flow

        :param attempts: used to limit max muber of attempts when current login fails
        :param verbose: Set to true for verbose output
        :return status_code:
        """
        if verbose: log.info("Starting: Login")

        # Get URL for Client
        if not self.url:
            if verbose: log.info("[1] Retrieving URL...")
            uri = f"https://{self.host}/saml/login-client/?get_url=True"
            request = requests.get(uri)
            if request.status_code==200:
                if verbose: log.info("[2] Success: Get URL")
                data = request.json()
                self.url = data.get("url")
                self.client_id = data.get("client_id")
            else:
                log.error(f"Failed {request.status_code}: Get URL")
                return request.status_code

        # Get Token after login
        webbrowser.open(self.url)
        if verbose: log.info("[5] Requesting token...")
        request = requests.get(f"https://{self.host}/saml/login-client/?client_id={self.client_id}")
        status_code = request.status_code

        if status_code==200:
            if verbose: log.info("Success: Login")
            self._save_token(request.json())
            return status_code
        elif status_code==408:
            if attempts<MAX_ATTEMPTS-1:
                if verbose: log.info("Trying again... [{}/{}]".format(attempts+1, MAX_ATTEMPTS-1))
                status_code = self._login(attempts=attempts+1, verbose=verbose)
            else:
                log.error(f"Failed {status_code}: Login")

        return status_code

    def _logout(self, verbose=False):
        """
        Log the client out

        :param verbose: Set to true for verbose output
        :return:
        """
        if verbose: log.info("Starting: Logout")
        data = self._get(f"https://{self.host}/saml/logout-client/") # Logout
        # if verbose: log.info("data: {}".format(data))
        resp = data.get("response")
        if resp:
            logout_url = resp.get("url")
            if logout_url: webbrowser.open(logout_url)
        self.clean_token() # Clean up Token
        return


    ############################################################################
    #                      Token-related methods                               #
    ############################################################################

    @property
    def is_token(self):
        return bool(self._token.get('token'))

    def clean_token(self):
        self.url = None
        self.client_id = None
        self._token = {
            'token': ''
        }

    def _save_token(self, token):
        # Set token for the future requests
        self._token = token
        self._token_time = datetime.now()

    def refresh_token(self):
        log.info("Starting: Refresh Token")

        # Refresh token
        request = requests.post(f"https://{self.host}/saml/refresh-token/", self._token)

        if request.status_code == 200:
            log.info("Success: Refreshing Token")
            self._save_token(request.json())
            return request.status_code

        log.error(f"Failed {request.status_code}: Refresh Token Expired")
        return request.status_code

    def _check_token(self):
        """
        Check if token is still valid, re-login into system if refresh token is expired

        :return:
        """
        if self._token_time and datetime.now() > self._token_time + timedelta(days=TOKEN_REFRESH):
            status = self.refresh_token()
            if status != 200:
                self.clean_token()
                self.login()
        return



"""
Copyright (c) 2020-2022 Go2Market Insights, LLC d/b/a Analyzr
All rights reserved.
https://analyzr.ai

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import requests
import json
from copy import deepcopy

from .constants import *


class BasicAuthClient:
    """
    The BasicAuthClient class manages low-level interactions with the Analyzr API
    using basic authentication (when applicable).

    """
    def __init__(self, user=USER, pwd=PWD, host=None):
        """
        :param user:
        :param pwd:
        :param host:
        :return:
        """
        if host is None:
            print('ERROR! Please provide a valid host, e.g. host=acme.api.g2m.ai')
            exit(1)
        self.__user = user
        self.__pwd = pwd
        self._base_url = 'https://{}/api/v1'.format(host)
        return

    ############################################################################
    #                           Basic API calls                                #
    ############################################################################

    def _get(self, uri, params=None):
        """
        Returns standard GET response for URI. Note that this
        function is protected, not private.
        :param uri:
        :param params:
        :return res:
        """
        if params is None:
            r = requests.get(uri, auth=(self.__user, self.__pwd))
        else:
            r = requests.get(uri, auth=(self.__user, self.__pwd), params=params)
        return self._response(r)

    def _post(self, uri, json_obj, compressed=False):
        """
        Returns standard POST response for URI. Note that this
        function is protected, not private.
        :param uri:
        :param json_obj:
        :param compressed: not used at this time
        :return res:
        """
        data = json.dumps(json_obj)
        r = requests.post(uri, auth=(self.__user, self.__pwd), json=data)
        return self._response(r)

    def _response(self, r):
        """
        """
        res = {}
        res['status'] = r.status_code
        if r.status_code>=200 and r.status_code<300:
            res['response'] = r.json()
        else:
            print('WARNING! Request returned status code: {}'.format(r.status_code))
            res['response'] = None
        return res



"""
Copyright (c) 2020-2021 Go2Market Insights, LLC
All rights reserved.
https://g2m.ai

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import uuid
import numpy as np
import pandas as pd
from copy import deepcopy

def xref_encode(series):
    """
    Replaces categorical values in a Series with UUIDs

    :param series:
    :return series2:
    :return xref:
    """
    # generate xref
    values = series.value_counts().index
    xref = {'forward': {np.nan: np.nan}, 'reverse': {np.nan: np.nan}}
    for val in values:
        key = str(uuid.uuid4())
        xref['forward'][val] = key
        xref['reverse'][key] = val

    # convert series
    series2 = xref_encode_with_keys(series, xref)
    return series2, xref

def xref_encode_with_keys(series, xref):
    """
    Replaces categorical values in a Series with UUIDs using existing encoding keys

    :param series:
    :param xref:
    :return series2:
    :return xref:
    """
    series2 = deepcopy(series)
    skipped_vals = []
    for idx, val in series2.iteritems():
        if val in xref['forward'].keys():
            series2[idx] = xref['forward'][val]
        else:
            series2[idx] = None
            if val not in skipped_vals: skipped_vals.append(val)
    if len(skipped_vals)>0:
        print('        WARNING! The following values were not present in the training encoding set and will be skipped: ', skipped_vals)
    return series2

def xref_decode(series, xref):
    """
    Replaces UUIDs in a Series with matching categorical values

    :param series:
    :param xref:
    :return series2:
    """
    series2 = deepcopy(series)
    for idx, val in series.iteritems():
        series2[idx] = xref['reverse'][val]
    return series2

def zref_encode(series):
    """
    Replaces numerical values in a Series with z scores

    :param series:
    :return series2:
    :return zref:
    """
    zref = { 'mean': series.mean(), 'stdev': series.std() }
    series2 = zref_encode_with_keys(series, zref)
    return series2, zref

def zref_encode_with_keys(series, zref):
    """
    Replaces numerical values in a Series with z scores using existing encoding keys

    :param series:
    :param zref:
    :return series2:
    """
    series2 = deepcopy(series)

    # Remove mean
    if np.isnan(zref['mean'])==False:
        series2 -= zref['mean']
    else:
        print('ERROR! series mean is nan, cannot encode series')
        return series, {}

    # Scale with std deviation
    if np.isnan(zref['stdev'])==False and zref['stdev']!=0.0:
        series2 /= zref['stdev']

    return series2

def zref_decode(series, zref):
    """
    Replaces z scores in a Series with denormalized numerical values

    :param series:
    :param zref:
    :return series2:
    """
    series2 = deepcopy(series)

    # Rescale with std deviation
    if np.isnan(zref['stdev'])==False and zref['stdev']!=0.0:
        series2 *= zref['stdev']

    # Add back mean
    if np.isnan(zref['mean'])==False:
        series2 += zref['mean']

    return series2

def bref_encode(series):
    """
    Boolean variables are not encoded at this stage

    :param series:
    :return series2:
    :return bref:
    """
    bref = {}
    series2 = bref_encode_with_keys(series, bref)
    return series2, bref

def bref_encode_with_keys(series, bref):
    """
    Boolean variables are not encoded at this stage

    :param series:
    :return series2:
    :return bref:
    """
    series2 = deepcopy(series)
    return series2

def bref_decode(series, bref):
    """
    Boolean variables are not encoded at this stage

    :param series:
    :param bref:
    :return series2:
    """
    series2 = deepcopy(series)
    return series2

def rref_encode(df, record_id_var):
    """
    Encode record ID column

    :param df:
    :param record_id_var:
    :return df2:
    :return rref:
    """
    df2 = deepcopy(df)
    df2[record_id_var], rref = xref_encode(df[record_id_var])
    return df2, rref

def rref_encode_with_keys(df, record_id_var, rref):
    """
    Encode record ID column using existing encoding keys

    :param df:
    :param record_id_var:
    :param rref:
    :return df2:
    """
    df2 = deepcopy(df)
    df2[record_id_var] = xref_encode_with_keys(df[record_id_var], rref)
    return df2

def rref_decode(df, record_id_var, rref):
    """
    :param df:
    :param record_id_var:
    :param rref:
    :return df2:
    """
    if rref=={}: return df
    df2 = deepcopy(df)
    if record_id_var not in df2.columns:
        df2[record_id_var] = df2.index
    df2[record_id_var] = xref_decode(df2[record_id_var], rref)
    df2 = df2.set_index(record_id_var)
    return df2

def fref_encode(df):
    """
    Encode field names

    :param df:
    :return df2:
    :return fref:
    """
    df2 = deepcopy(df)
    fref = {'forward': {'PC_ID': 'PC_ID'}, 'reverse': {'PC_ID': 'PC_ID'}}
    counter = 0
    cols2 = []
    for col in df2.columns:
        key = 'X_{}'.format(counter)
        fref['forward'][col] = key
        fref['reverse'][key] = col
        cols2.append(key)
        counter += 1
    df2.columns = cols2
    return df2, fref

def fref_encode_with_keys(df, fref):
    """
    Encode field names using existing encoding keys

    :param df:
    :param fref:
    :return df2:
    """
    df2 = deepcopy(df)
    df2.columns = [ fref['forward'][col] for col in df2.columns ]
    return df2

def fref_to_fref_expanded(fref, xref):
    """
    Convert <fref> to expanded field names (using dummy variable convention)

    :param fref:
    :param xref:
    :return fref_exp:
    """
    # print('[fref_to_fref_expanded] fref: ', fref)
    # print('[fref_to_fref_expanded] xref: ', xref)
    fref_exp = {'forward': {}, 'reverse': {}}
    for col in fref['forward'].keys():
        key = fref['forward'][col]
        if col in xref.keys():
            # col is categorical, need to expand it
            for category in xref[col]['forward'].keys():
                col_exp = '{}_{}'.format(col, category)
                key_exp = '{}_{}'.format(key, xref[col]['forward'][category])
                fref_exp['forward'][col_exp] = key_exp
                fref_exp['reverse'][key_exp] = col_exp
        else:
            # col is not categorical
            fref_exp['forward'][col] = key
            fref_exp['reverse'][key] = col
    # print('[fref_to_fref_expanded] fref_exp: ', fref_exp)
    return fref_exp

def fref_decode(df, fref):
    """
    Decode field names

    :param df:
    :param fref:
    :return df2:
    """
    if fref=={}: return df
    df2 = deepcopy(df)
    # df2.columns = [ fref['reverse'][col] for col in df2.columns ]
    df2.columns = fref_decode_columns(df2.columns, fref)
    return df2

def fref_decode_columns(cols, fref):
    """
    Decode array of column names

    :param cols:
    :param fref:
    :return cols2:
    """
    if fref=={}: return cols
    cols2 = [ fref_decode_value(col, fref) for col in cols ]
    return cols2

def fref_decode_value(col, fref):
    """
    Decode single column name

    :param col:
    :param fref:
    :return col2:
    """
    col2 = col
    if fref!={} and col in fref['reverse'].keys():
        col2 =  fref['reverse'][col]
    return col2

def compute_cluster_stats(df, categorical_fields):
    """
    :param df:
    :param categorical_fields:
    :return stats:
    """
    if df is None: return None
    stats = pd.DataFrame(df['PC_ID'].value_counts())
    stats.columns=['count']
    stats['frequency'] = stats['count']/len(df)
    stats.sort_index(inplace=True)
    stats.index.rename('PC_ID', inplace=True)
    stats = stats.join(pd.get_dummies(df, columns=categorical_fields).groupby('PC_ID').mean(), on='PC_ID')
    return stats.T

def compute_cluster_distances(df, categorical_fields):
    """
    :param df:
    :param categorical_fields:
    :return distances:
    """
    if df is None: return None
    df2 = pd.get_dummies(df, columns=categorical_fields).groupby('PC_ID').mean()
    recs = []
    for idx1, cluster1 in df2.iterrows():
        rec = []
        for idx2, cluster2 in df2.iterrows():
            rec.append(np.linalg.norm(cluster1-cluster2))
        recs.append(rec)
    return pd.DataFrame.from_records(recs)

def merge_cluster_ids(df, pc_id, idx_var):
    """
    Merge principal component IDs back onto original records and compute stats by cluster

    :param df: original dataframe to be clustered
    :param pc_id: principal component IDs by record
    :param idx_var: name of record ID field
    :return df2:
    """
    return pd.merge(df, pc_id, left_on=idx_var, right_on=idx_var, how='left')

def get_test_data():
    """
    :param:
    :return df:
    """
    df = pd.read_csv('https://g2mstaticfiles.blob.core.windows.net/$web/titanic.csv', encoding = "ISO-8859-1", low_memory=False)
    return df



import os, sys, time, json
import pandas as pd
from copy import deepcopy
from io import StringIO
import pickle

from .constants import *
from .utils import *

DEBUG = False

class BaseRunner:
    """
    Base class for all task runners

    """
    def __init__(self, client=None, base_url=None):
        """
        :param client: client object from SamlSsoAuthClient class
        :return:
        """
        self._client = client
        self._base_url = base_url
        return

    def _get_request_id(self):
        """
        :param:
        :return uuid:
        """
        return str(uuid.uuid4())

    def _buffer_save(self, df, batch_size=1000, client_id=None, request_id=None, verbose=False, compressed=False, staging=False):
        """
        :param df:
        :param batch_size:
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param request_id:
        :param verbose: Set to true for verbose output
        :param compressed:
        :param staging:
        :return res:
        """
        if verbose: print('Saving data to buffer...')
        if request_id is None:
            print('WARNING! Request ID is None, aborting buffer save. ')
            return {'request_id': None, 'total_batches': None, 'batches_saved': None}
        batched_df = [ df[i:i+batch_size] for i in range(0, len(df), batch_size) ]
        idx = 0
        success = 0
        for batch in batched_df:
            idx += 1
            for attempt in range(0, 5):
                if self._batch_save(batch, idx, len(batched_df), client_id, request_id, verbose=(attempt==4), compressed=compressed, staging=staging):
                    if verbose: sys.stdout.write('        Processed batch {} of {}\r'.format(idx, len(batched_df)))
                    success += 1
                    break
                else:
                    time.sleep(2)
        if verbose: sys.stdout.write('\n')
        return {'request_id': request_id, 'total_batches': len(batched_df), 'batches_saved': success}

    def _batch_save(self, batch, idx, n, client_id, request_id, verbose=False, compressed=False, staging=False):
        """
        If staging is false, converts dataframe to dictionary and uploads to database buffer.
        If staging is true, converts dataframe to CSV chunk and uploads to data lake buffer.

        :param batch:
        :param idx:
        :param n:
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param request_id:
        :param verbose: Set to true for verbose output
        :param compressed:
        :return success:
        """
        success = False
        uri = '{}/buffer/'.format(self._base_url)
        # if DEBUG: print('[_batch_save] batch #{} len: {} staging: {}'.format(idx, len(batch), staging))
        try:
            res = self._client._post(
                uri,
                {
                    'command': 'upload',
                    'data': batch.to_dict() if not staging else batch.to_csv(header=(idx==1), index=False),
                    'request_id': request_id,
                    'client_id': client_id,
                    'staging': staging,
                },
                compressed=compressed,
            )
            if res['status']==200:
                success = True
            else:
                if verbose: print('[__batch_save] WARNING! Returned status {} for batch {} of {}'.format(res['status'], idx, n))
        except:
            if verbose: print('[__batch_save] WARNING! API call failed for batch {} of {}'.format(idx, n))
        return success

    def _buffer_read(self, client_id=None, request_id=None, dataframe_name='df', verbose=False, raw=False, staging=False):
        """
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param request_id:
        :param verbose: Set to true for verbose output
        :param raw:
        :param staging:
        :return df:
        """
        # if DEBUG: print('[_buffer_read] making call to read buffer')
        res2 = self.__read(client_id=client_id, request_id=request_id, dataframe_name=dataframe_name, verbose=verbose, staging=staging)
        # if DEBUG: print('[_buffer_read] Buffer returned response')
        if raw: return res2
        if res2['status']!=200:
            print('ERROR! Buffer read failed: {}'.format(res2))
            return pd.DataFrame(None)
        # if DEBUG: print('[_buffer_read] Buffer read successfully')
        if res2['response']['data'] is None: return pd.DataFrame()
        return pd.DataFrame(res2['response']['data']) if not staging else pd.read_csv(StringIO(res2['response']['data']))

    def __read(self, client_id=None, request_id=None, dataframe_name='df', verbose=False, staging=False):
        """
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param request_id:
        :param verbose: Set to true for verbose output
        :param staging:
        :return res:
        """
        uri = '{}/buffer/'.format(self._base_url)
        return self._client._post(uri, {
            'command': 'read',
            'request_id': request_id,
            'client_id': client_id,
            'dataframe_name': dataframe_name,
            'staging': staging,
        })

    def queue_purge(self, client_id=None, verbose=False):
        """
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param verbose: Set to true for verbose output
        :return res:
        """
        if verbose: print('Purging queue...')
        uri = '{}/analytics/'.format(self._base_url)
        res = self._client._post(uri, {
            'command': 'purge',
            'client_id': client_id,
        })
        if res['status']!=200: print('WARNING! Queue purge failed: {}'.format(res))
        return res

    def buffer_purge(self, client_id=None, verbose=False):
        """
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param verbose: Set to true for verbose output
        :return res:
        """
        if verbose: print('Purging buffer...')
        uri = '{}/buffer/'.format(self._base_url)
        res = self._client._post(uri, {
            'command': 'purge',
            'client_id': client_id,
        })
        if res['status']!=200: print('WARNING! Buffer purge failed: {}'.format(res))
        return res

    def buffer_usage(self, client_id=None, verbose=False):
        """
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param verbose: Set to true for verbose output
        :return res:
        """
        if verbose: print('Checking buffer usage...')
        uri = '{}/buffer/'.format(self._base_url)
        res = self._client._post(uri, {
            'command': 'usage',
            'client_id': client_id,
        })
        if res['status']!=200: print('WARNING! Checking buffer usage failed: {}'.format(res))
        return res

    def _buffer_clear(self, client_id=None, request_id=None, verbose=False):
        """
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param request_id:
        :param verbose: Set to true for verbose output
        :param staging:
        :return res:
        """
        if verbose: print('Clearing buffer...')
        uri = '{}/buffer/'.format(self._base_url)
        res = self._client._post(uri, {
            'command': 'clear',
            'request_id': request_id,
            'client_id': client_id,
        })
        if res['status']!=200: print('WARNING! Buffer clear failed: {}'.format(res))
        return res

    def _encode(self, df, keys=None, categorical_vars=[], numerical_vars=[], bool_vars=[], record_id_var=None, encode_field_names=True, verbose=False):
        """
        Encode dataframe for the categorical and numerical variables provided.

        :param df:
        :param keys:
        :param categorical_vars:
        :param numerical_vars:
        :param record_id_var:
        :param encode_field_names:
        :param verbose: Set to true for verbose output
        :return df2:
        :return xref: cross-reference dictionary for categorical variables
        :return zref: cross-reference dictionary for numerical variables
        :return rref: cross-reference dictionary for the record ID variable
        :return fref: cross-reference dictionary for field names
        """
        return self._encode_no_keys(
            df,
            categorical_vars=categorical_vars, numerical_vars=numerical_vars, bool_vars=bool_vars, record_id_var=record_id_var,
            encode_field_names=encode_field_names, verbose=verbose,
        ) if keys is None else self._encode_with_keys(
            df, keys,
            categorical_vars=categorical_vars, numerical_vars=numerical_vars, bool_vars=bool_vars, record_id_var=record_id_var,
            encode_field_names=encode_field_names, verbose=verbose,
        )

    def _encode_no_keys(self, df, categorical_vars=[], numerical_vars=[], bool_vars=[], record_id_var=None, encode_field_names=True, verbose=False):
        """
        Encode dataframe for the categorical and numerical variables provided. This
        assumes no pre-existing encoding, returns encoded data and matching
        encoding keys.

        :param df:
        :param categorical_vars:
        :param numerical_vars:
        :param record_id_var:
        :param encode_field_names:
        :param verbose: Set to true for verbose output
        :return df2:
        :return xref: cross-reference dictionary for categorical variables
        :return zref: cross-reference dictionary for numerical variables
        :return rref: cross-reference dictionary for the record ID variable
        :return fref: cross-reference dictionary for field names
        """
        df2 = deepcopy(df)

        # Encode categorical variables
        if verbose: print('Encoding categorical variables:')
        xref = {}
        for col in categorical_vars:
            if verbose: print('\t{}'.format(col))
            df2[col], xref[col] = xref_encode(df[col])

        # Encode numerical variables
        if verbose: print('Encoding numerical variables:')
        zref = {}
        for col in numerical_vars:
            if verbose: print('\t{}'.format(col))
            df2[col], zref[col] = zref_encode(df[col])

        # Encode boolean variables
        # if verbose: print('Encoding boolean variables:')
        bref = {}
        for col in bool_vars:
            if verbose: print('\t{}'.format(col))
            df2[col], bref[col] = bref_encode(df[col])

        # Encode record ID
        if record_id_var is not None:
            if verbose: print('Encoding record IDs...')
            df2, rref = rref_encode(df2, record_id_var)
        else:
            rref = {}

        # Encode field names
        if encode_field_names:
            if verbose: print('Encoding field names...')
            df2, fref = fref_encode(df2)
            fref_exp = fref_to_fref_expanded(fref, xref)

        return df2, xref, zref, rref, fref, fref_exp, bref

    def _encode_with_keys(self, df, keys, categorical_vars=[], numerical_vars=[], bool_vars=[], record_id_var=None, encode_field_names=True, verbose=False):
        """
        Encode dataframe for the categorical and numerical variables provided. This
        assumes pre-existing encoding keys are provided, returns encoded data and
        encoding keys.

        Note: record IDs are re-encoded every time with new keys to accommodate
        additional records.

        :param df:
        :param keys:
        :param categorical_vars:
        :param numerical_vars:
        :param record_id_var:
        :param encode_field_names:
        :param verbose: Set to true for verbose output
        :return df2:
        :return xref: cross-reference dictionary for categorical variables
        :return zref: cross-reference dictionary for numerical variables
        :return rref: cross-reference dictionary for the record ID variable
        :return fref: cross-reference dictionary for field names
        """
        df2 = deepcopy(df)

        # Encode categorical variables
        if verbose: print('Encoding categorical variables:')
        xref = keys['xref']
        for col in categorical_vars:
            if verbose: print('\t{}'.format(col))
            df2[col] = xref_encode_with_keys(df[col], xref[col])

        # Encode numerical variables
        if verbose: print('Encoding numerical variables:')
        zref = keys['zref']
        for col in numerical_vars:
            if verbose: print('\t{}'.format(col))
            df2[col] = zref_encode_with_keys(df[col], zref[col])

        # Encode boolean variables
        # if verbose: print('Encoding boolean variables:')
        bref = keys['bref']
        for col in bool_vars:
            if verbose: print('\t{}'.format(col))
            df2[col] = bref_encode_with_keys(df[col], bref[col])

        # Encode record ID
        if record_id_var is not None:
            if verbose: print('Encoding record IDs...')
            # rref = keys['rref']
            # df2 = rref_encode_with_keys(df2, record_id_var, rref)
            df2, rref = rref_encode(df2, record_id_var)
        else:
            rref = {}

        # Encode field names
        if encode_field_names:
            if verbose: print('Encoding field names...')
            fref = keys['fref']
            fref_exp = keys['fref_exp']
            df2 = fref_encode_with_keys(df2, fref)
        else:
            fref = {}
            fref_exp = {}

        df2 = df2.dropna()

        return df2, xref, zref, rref, fref, fref_exp, bref

    def _decode(self, df, categorical_vars=[], numerical_vars=[], bool_vars=[], record_id_var=None, xref={}, zref={}, rref={}, fref={}, bref={}, verbose=False):
        """
        Decode dataframe using the encoding keys provided (xref, zrfef, rref, fref)

        :param df:
        :param categorical_vars:
        :param numerical_vars:
        :param record_id_var:
        :param xref: cross-reference dictionary for categorical variables
        :param zref: cross-reference dictionary for numerical variables
        :param rref: cross-reference dictionary for the record ID variable
        :param fref: cross-reference dictionary for field names
        :param verbose: Set to true for verbose output
        :return df2:
        """
        if df is None or df.empty: return df
        df2 = deepcopy(df)

        # Decode field names
        if verbose: print('Decoding field names...')
        df2 = fref_decode(df2, fref) # decode field names first

        # Decode categorical variables
        if verbose: print('Decoding categorical variables:')
        for col in categorical_vars:
            if col in df2.columns:
                if verbose: print('\t{}'.format(col))
                df2[col] = xref_decode(df2[col], xref[col])

        # Decode numerical variables
        if verbose: print('Decoding numerical variables:')
        for col in numerical_vars:
            if col in df2.columns:
                if verbose: print('\t{}'.format(col))
                df2[col] = zref_decode(df2[col].astype('float'), zref[col])

        # Decode boolean variables
        # if verbose: print('Decoding boolean variables:')
        for col in bool_vars:
            if col in df2.columns:
                if verbose: print('\t{}'.format(col))
                df2[col] = bref_decode(df2[col].astype('int'), zref[col])

        # Decode record ID
        if record_id_var is not None and record_id_var in df2.columns:
            if verbose: print('Decoding record IDs...')
            df2 = rref_decode(df2, record_id_var, rref)

        return df2

    def _poll(self, payload={}, timeout=600, step=1, verbose=False):
        """
        :param payload:
        :param timeout:
        :param step:
        :param verbose: Set to true for verbose output
        :return res:
        """
        counter = 0
        res = {}
        uri = '{}/analytics/'.format(self._base_url)
        while counter<timeout:
            time.sleep(step)
            try:
                res = self._client._post(uri, payload)
                if verbose: sys.stdout.write('[_poll][{}] {}\r'.format(counter, res))
                if res['response']['status'] in ['Complete', 'Failed'] or 'Failed:' in res['response']['status']: break
            except:
                pass
            counter += step
        if verbose: sys.stdout.write('\n')
        return res

    def _status(self, request_id=None, client_id=None, verbose=False):
        """
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param verbose: Set to true for verbose output
        :return res:
        """
        res = {}
        uri = '{}/analytics/'.format(self._base_url)
        if request_id is None or client_id is None:
            print('[_status] ERROR! Invalid status request (request_id: {}, client_id: {})'.format(request_id, client_id))
        else:
            try:
                res = self._client._post(uri, {'request_id': request_id, 'client_id': client_id, 'command': 'task-status'})
            except:
                if verbose: print('[_status] ERROR! Could not retrieve status for request ID: {} and client ID: {}'.format(request_id, client_id))
            else:
                if res['status']==200:
                    res = res['response']
                else:
                    if verbose: print('[_status] ERROR! Status inquiry for request ID: {} and client ID: {} returned status: {}'.format(request_id, client_id, res['status']))
                    res = {}
        return res

    def _keys_save(self, model_id=None, keys=None, verbose=False):
        """
        :param model_id:
        :param keys:
        :param verbose: Set to true for verbose output
        :return:
        """
        if verbose: print('Saving encoding keys locally...')
        if not os.path.exists(TEMP_DIR): os.makedirs(TEMP_DIR)
        filename = '{}/{}.bin'.format(TEMP_DIR, model_id)
        file = open(filename, 'wb')
        pickle.dump(keys, file)
        file.close()
        return

    def _keys_load(self, model_id=None, verbose=False):
        """
        :param model_id:
        :param verbose: Set to true for verbose output
        :return keys:
        """
        if verbose: print('Loading encoding keys...')
        try:
            filename = '{}/{}.bin'.format(TEMP_DIR, model_id)
            file = open(filename, 'rb')
            keys = pickle.load(file)
            file.close()
        except FileNotFoundError:
            print('ERROR! Keys not found for model_id: {}'.format(model_id))
            keys = None

        return keys



import os, sys, time, json
import pandas as pd
from copy import deepcopy

from .runner_base import BaseRunner
from .constants import *
from .utils import *

class PropensityRunner(BaseRunner):
    """
    Run the propensity scoring pipeline

    :param client: SAML SSO client object
    :type client: SamlSsoAuthClient, required
    :param base_url: Base URL for the Analyzr API tenant
    :type base_url: str, required
    """
    def __init__(self, client=None, base_url=None):
        """
        """
        super().__init__(client=client, base_url=base_url)
        self.__uri = '{}/analytics/'.format(self._base_url)
        return

    def predict(
            self, df, model_id=None, client_id=None,
            idx_var=None, categorical_vars=[], numerical_vars=[], bool_vars=[],
            buffer_batch_size=1000, api_batch_size=2000, verbose=False,
            timeout=600, step=2, compressed=False, staging=True, encoding=True):
        """
        Predict probabilities of outcome (propensities) for user-specified
        dataset using a pre-trained model

        :param df: Dataframe containing dataset to be used for training. The
            data is homomorphically encrypted by the client prior to being
            transferred to the API buffer when `encoding` is set to `True`
        :type df: DataFrame, required
        :param model_id: UUID for a specific model object. Refers to a model
            that was previously trained
        :type model_id: str, required
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :type client_id: string, required
        :param idx_var: Name of index field identifying unique record IDs in
            `df` for audit purposes
        :type idx_var: string, required
        :param categorical_vars: Array of field names identifying categorical
            fields in the dataframe `df`
        :type categorical_vars: string[], required
        :param numerical_vars: Array of field names identifying categorical
            fields in the dataframe `df`
        :type numerical_vars: string[], required
        :param bool_vars: Array of field names identifying boolean fields in the
            dataframe `df`
        :type bool_vars: string[], optional
        :param buffer_batch_size: Batch size for the purpose of uploading data
            from the client to the server's buffer
        :type buffer_batch_size: int, optional
        :param api_batch_size: Batch size for the purpose of processing data in
            the API
        :type api_batch_size: int, optional
        :param verbose: Set to true for verbose output
        :type verbose: boolean, optional
        :param timeout: Client will keep polling API for a period of `timeout`
            seconds
        :type timeout: int, optional
        :param step: Polling interval, in seconds
        :type step: int, optional
        :param compressed: Perform additional compression when uploading data to
            buffer
        :type compressed: boolean, optional
        :param staging: When set to True the API will use temporay secure cloud
            storage to buffer the data rather than a relational database
            (default is `True`)
        :type staging: boolean, optional
        :param encoding: Encode and decode data with homomorphic encryption
        :type encoding: boolean, optional
        :return: JSON object with the following attributes:
                    `model_id` (UUID provided with initial request),
                    `data2`: original dataset with cluster IDs appended
        """
        res = {}
        res['model_id'] = model_id
        res['data2'] = pd.DataFrame()
        if api_batch_size<buffer_batch_size:
            api_batch_size = buffer_batch_size
            if verbose: print('Setting API_BATCH_SIZE to {:,}'.format(api_batch_size))
        batched_df = [ df[i:i+api_batch_size] for i in range(0, len(df), api_batch_size) ]
        idx = 1
        for batch in batched_df:
            if verbose: print('Processing API request {} of {}'.format(idx, len(batched_df)))
            res5 = self.__predict_api_batch(
                batch, model_id=model_id,
                client_id=client_id,
                idx_var=idx_var,
                categorical_vars=categorical_vars,
                numerical_vars=numerical_vars,
                bool_vars=bool_vars,
                buffer_batch_size=buffer_batch_size,
                verbose=verbose,
                timeout=timeout,
                step=step,
                compressed=compressed,
                staging=staging,
                encoding=encoding,
            )
            res['data2'] = res['data2'].append(res5['data2'])
            idx += 1
        return res

    def __predict_api_batch(
            self, df, model_id=None, client_id=None, idx_var=None,
            categorical_vars=[], numerical_vars=[], bool_vars=[],
            buffer_batch_size=1000, verbose=False, timeout=600, step=2,
            compressed=False, staging=False, encoding=True):
        """
        :param df:
        :param model_id:
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :param idx_var:
        :param categorical_vars:
        :param numerical_vars:
        :param buffer_batch_size: Batch size for the purpose of uploading data
            from the client to the server's buffer
        :param verbose: Set to true for verbose output
        :param compressed:
        :param staging:
        :param encoding:
        :return res5:
        """
        request_id = self._get_request_id()

        if encoding:
            keys = self._keys_load(model_id=model_id, verbose=verbose) # Load encoding keys
            if keys is None:
                print('ERROR! Keys not found. ')
                return None
            data, xref, zref, rref, fref, fref_exp, bref = self._encode(
                df, keys=keys, categorical_vars=categorical_vars,
                numerical_vars=numerical_vars, bool_vars=bool_vars,
                record_id_var=idx_var, verbose=verbose)
            if verbose: print('Total rows encoded: {:,}'.format(len(data)))
        else:
            data = df

        # Save data to buffer
        res = self._buffer_save(
            data,
            client_id=client_id,
            request_id=request_id,
            verbose=verbose,
            batch_size=buffer_batch_size,
            compressed=compressed,
            staging=staging
        )

        # Predict with propensity model and retrieve results
        if res['batches_saved']==res['total_batches']:
            self.__predict(
                request_id=res['request_id'],
                model_id=model_id,
                client_id=client_id,
                idx_field=fref['forward'][idx_var] if encoding else idx_var,
                categorical_fields=[ fref['forward'][var] for var in categorical_vars ] if encoding else categorical_vars,
                verbose=verbose,
                staging=staging
            )
            self._poll(
                payload={
                    'request_id': res['request_id'],
                    'client_id': client_id,
                    'command': 'task-status'
                    },
                timeout=timeout,
                step=step,
                verbose=verbose)
            data2 = self.__retrieve_predict_results(
                request_id=request_id,
                client_id=client_id,
                verbose=verbose)
        else:
            print('ERROR! Buffer save failed: {}'.format(res))

        # Clear buffer
        res4 = self._buffer_clear(
            request_id=res['request_id'], client_id=client_id,
            verbose=verbose)

        # Decode data
        if encoding:
            data2 = self._decode(
                data2,
                categorical_vars=categorical_vars,
                numerical_vars=numerical_vars,
                bool_vars=bool_vars,
                record_id_var=idx_var,
                xref=xref,
                zref=zref,
                rref=rref,
                fref=fref,
                bref=bref,
                verbose=verbose
            )

        # Compile results
        res5 = {}
        res5['data2'] = data2
        res5['model_id'] = model_id
        return res5


    def __predict(
            self, request_id=None, model_id=None, client_id=None,
            idx_field=None, categorical_fields=[], verbose=False,
            staging=False):
        """
        :param request_id:
        :param model_id:
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :param idx_field: Name of index field identifying unique record IDs for
            audit purposes
        :param categorical_fields:
        :param verbose: Set to true for verbose output
        :param staging:
        :return:
        """
        if verbose: print('Predicting propensity model using data in buffer...')
        res = self._client._post(self.__uri, {
            'command': 'propensity-predict',
            'model_id': model_id,
            'request_id': request_id,
            'client_id': client_id,
            'idx_field': idx_field,
            'categorical_fields': categorical_fields,
            'staging': staging,
        })
        return res

    def __retrieve_predict_results(
            self, request_id=None, client_id=None,
            verbose=False):
        """
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :param verbose: Set to true for verbose output
        :return data2:
        """
        data2 = self._buffer_read(
            request_id=request_id, client_id=client_id, dataframe_name='res',
            verbose=verbose, staging=True)
        return data2

    def check_status(
            self, model_id=None, client_id=None, verbose=False,
            encoding=True):
        """
        Check the status of a specific model run, and retrieve results if model
        run is complete

        :param model_id: UUID for a specific model object
        :type model_id: str, required
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :type client_id: string, required
        :param verbose: Set to true for verbose output
        :type verbose: boolean, optional
        :param encoding: Decode results with homomorphic encryption
        :type encoding: boolean, optional
        :return: JSON object with the following attributes, as applicable:
                    `status` (can be Pending, Complete, or Failed),
                    `features` (table of feature importances),
                    `confusion_matrix` (confusion matrix using test dataset),
                    `stats` (error stats including accuracy, precision, recall,
                    F1, AUC, Gini),
                    `roc` (receiver operating characteristic curve)
        """
        res1 = {}
        res1['model_id'] = model_id
        res2 = self._status(
            request_id=model_id, client_id=client_id,
            verbose=verbose)
        if res2!={} and 'status' in res2.keys():
            res1['status'] = res2['status']
            if res2['status']=='Complete':
                if encoding:
                    # Load encoding keys
                    keys = self._keys_load(model_id=model_id, verbose=verbose)
                    if keys is None:
                        print('ERROR! Keys not found. ')
                        return None
                else:
                    keys = {}
                features, confusion_matrix, stats, roc = self.__retrieve_train_results(
                        request_id=model_id,
                        client_id=client_id,
                        fref=keys['fref_exp'] if encoding else {},
                        verbose=verbose,
                        encoding=encoding)
                res1['features'] = features
                res1['confusion_matrix'] = confusion_matrix
                res1['stats'] = stats
                res1['roc'] = roc
            if res2['status'] in ['Complete', 'Failed']:
                self._buffer_clear(
                    request_id=model_id, client_id=client_id,
                    verbose=verbose)
        return res1

    def train(self, df, client_id=None,
            idx_var=None, outcome_var=None, categorical_vars=[], numerical_vars=[],
            bool_vars=[], algorithm='random-forest-classifier', train_size=0.5,
            buffer_batch_size=1000, verbose=False, timeout=600, step=2, poll=True,
            smote=False, param_grid=None, scoring=None, n_splits=None,
            compressed=False, staging=True, encoding=True):
        """
        Train propensity model on user-provided dataset

        :param df: Dataframe containing dataset to be used for training. The data
            is homomorphically encrypted by the client prior to being transferred
            to the API buffer when `encoding` is set to `True`
        :type df: DataFrame, required
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :type client_id: string, required
        :param idx_var: Name of index field identifying unique record IDs in
            `df` for audit purposes
        :type idx_var: string, required
        :param outcome_var: Name of dependent variable, usually a boolean
            variable set to `0` or `1`
        :type outcome_var: string, required
        :param categorical_vars: Array of field names identifying categorical
            fields in the dataframe `df`
        :type categorical_vars: string[], required
        :param numerical_vars: Array of field names identifying categorical
            fields in the dataframe `df`
        :type numerical_vars: string[], required
        :param bool_vars: Array of field names identifying boolean fields in
            the dataframe `df`
        :type bool_vars: string[], optional
        :param algorithm: Can be any of the following:
            `random-forest-classifier`, `gradient-boosting-classifier`,
            `xgboost-classifier`, `ada-boost-classifier`,
            `extra-trees-classifier`, `logistic-regression-classifier`.
            Algorithms are sourced from Scikit-Learn unless otherwise indicated.
            Additional algorithms may be available
        :type algorithm: string, required
        :param train_size: Share of training dataset assigned to training vs.
            testing, e.g. if train_size is set to 0.8 80% of the dataset will be
            assigned to training and 20% will be randomly set aside for testing
            and validation
        :type train_size: float, optional
        :param buffer_batch_size: Batch size for the purpose of uploading data
            from the client to the server's buffer
        :type buffer_batch_size: int, optional
        :param verbose: Set to true for verbose output
        :type verbose: boolean, optional
        :param timeout: Client will keep polling API for a period of `timeout`
            seconds
        :type timeout: int, optional
        :param step: Polling interval, in seconds
        :type step: int, optional
        :param poll: Keep polling API while the job is being run (default is
            `True`)
        :type poll: boolean, optional
        :param smote: Apply SMOTE pre-processing
        :type smote: boolean, optional
        :param param_grid: Parameter grid to be used during the cross-validation
            grid search (hypertuning). The default is algorithm-specific and set
            by the API.
        :type param_grid: JSON object, optional
        :param scoring: Scoring methodology to evaluate the performance of the
            cross-validated model. Common methodologies include `roc_auc`,
            `accuracy`, and `f1`. Default is algorithm-specific and set by the
            API
        :type scoring: string, optional
        :param n_splits: Number of folds. Must be at least 2, defaults to 10
        :type n_splits: int, optional
        :param compressed: Perform additional compression when uploading data to
            buffer. Defaults to `False`
        :type compressed: boolean, optional
        :param staging: When set to True the API will use temporay secure cloud
            storage to buffer the data rather than a relational database.
            Defaults to `True`
        :type staging: boolean, optional
        :param encoding: encode and decode data with homomorphic encryption.
            Defaults to `True`
        :type encoding: boolean, optional
        :return: JSON object with the following attributes, as applicable:
                    `model_id` (UUID provided with initial request),
                    `features` (table of feature importances),
                    `confusion_matrix` (confusion matrix using test dataset),
                    `stats` (error stats including accuracy, precision, recall,
                    F1, AUC, Gini),
                    `roc` (receiver operating characteristic curve)
        """

        # Encode data
        request_id = self._get_request_id()
        if verbose: print('Model ID: {}'.format(request_id))
        if encoding:
            data, xref, zref, rref, fref, fref_exp, bref = self._encode(
                df, categorical_vars=categorical_vars,
                numerical_vars=numerical_vars, bool_vars=bool_vars,
                record_id_var=idx_var, verbose=verbose)
            self._keys_save(
                model_id=request_id,
                keys={
                    'xref': xref,
                    'zref': zref,
                    'rref': rref,
                    'fref': fref,
                    'fref_exp': fref_exp,
                    'bref': bref
                },
                verbose=verbose) # Save encoding keys locally
        else:
            data = df

        # Save encoded data to buffer
        res = self._buffer_save(
            data, client_id=client_id, request_id=request_id, verbose=verbose,
            batch_size=buffer_batch_size, compressed=compressed, staging=staging)

        # Train propensity model and retrieve results
        if res['batches_saved']==res['total_batches']:
            self.__train(
                request_id=res['request_id'],
                client_id=client_id,
                idx_field=fref['forward'][idx_var] if encoding else idx_var,
                outcome_var=fref['forward'][outcome_var] if encoding else outcome_var,
                categorical_fields=[ fref['forward'][var] for var in categorical_vars ] if encoding else categorical_vars,
                algorithm=algorithm,
                train_size=train_size,
                smote=smote,
                verbose=verbose,
                staging=staging,
                param_grid=param_grid,
                scoring=scoring,
                n_splits=n_splits,
            )
            if poll:
                res2 = self._poll(
                    payload={
                        'request_id': res['request_id'],
                        'client_id': client_id,
                        'command': 'task-status'
                    },
                    timeout=timeout,
                    step=step,
                    verbose=verbose)
                if res2['response']['status'] in ['Complete']:
                    features, confusion_matrix, stats, roc = self.__retrieve_train_results(
                        request_id=request_id,
                        client_id=client_id,
                        fref=fref_exp if encoding else {},
                        verbose=verbose,
                        encoding=encoding)
                else:
                    print('WARNING! Training request came back with status: {}'.format(res2['response']['status']))
        else:
            print('ERROR! Buffer save failed: {}'.format(res))

        # Clear buffer
        if poll: res4 = self._buffer_clear(
            request_id=res['request_id'], client_id=client_id,
                verbose=verbose)

        # Compile results
        if verbose and not poll: print('Training job started with polling disabled. You will need to request results for this model ID.')
        res5 = {}
        res5['model_id'] = request_id
        if poll:
            res5['features'] = features
            res5['confusion_matrix'] = confusion_matrix
            res5['stats'] = stats
            res5['roc'] = roc
        return res5

    def __train(self, request_id=None, client_id=None,
            idx_field=None, outcome_var=None, categorical_fields=[],
            algorithm='random-forest-classifier', train_size=0.5, smote=False,
            param_grid=None, scoring=None, n_splits=None,
            verbose=False, staging=False):
        """
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :param idx_field: Name of index field identifying unique record IDs for
            audit purposes
        :param outcome_var:
        :param categorical_fields:
        :param algorithm:
        :param train_size:
        :param smote:
        :param verbose: Set to true for verbose output
        :param staging:
        :param param_grid:
        :param scoring:
        :param n_splits:
        :return:
        """
        if verbose: print('Training propensity model using data in buffer...')
        res = self._client._post(self.__uri, {
            'command': 'propensity-train',
            'request_id': request_id,
            'client_id': client_id,
            'algorithm': algorithm,
            'train_size': train_size,
            'smote': smote,
            'idx_field': idx_field,
            'outcome_var': outcome_var,
            'categorical_fields': categorical_fields,
            'staging': staging,
            'param_grid': param_grid,
            'scoring': scoring,
            'n_splits': n_splits,
        })
        if verbose: print('Training request posted.')
        return res

    def __retrieve_train_results(
            self, request_id=None, client_id=None, fref={}, verbose=False,
            encoding=True):
        """
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :param fref:
        :param verbose: Set to true for verbose output
        :param encoding:
        :return features:
        :return confusion_matrix:
        :return stats:
        """
        if verbose: print('Retrieving training results...')

        # Features
        if verbose: print('    Retrieving features...')
        features = self._buffer_read(
            request_id=request_id, client_id=client_id, dataframe_name='features',
            verbose=verbose)
        if not features.empty:
            features['Importance'] = features['Importance'].astype('float')
            features.sort_values(by=['Importance'], ascending=False, inplace=True)
            if encoding:
                for idx, row in features.iterrows():
                    features.loc[idx, 'Feature'] = fref_decode_value(
                        features.loc[idx, 'Feature'],
                        fref)
        else:
            if verbose: print('WARNING! Features dataframe is empty')

        # Confusion matrix
        if verbose: print('    Retrieving confusion matrix...')
        confusion_matrix = self._buffer_read(
            request_id=request_id, client_id=client_id,
            dataframe_name='confusion_matrix', verbose=verbose)

        # Stats
        if verbose: print('    Retrieving performance stats...')
        stats = self._buffer_read(
            request_id=request_id, client_id=client_id, dataframe_name='stats',
            verbose=verbose)
        if not stats.empty:
            stats['Value'] = stats['Value'].astype('float')
        else:
            if verbose: print('WARNING! Stats dataframe is empty')

        # ROC
        if verbose: print('    Retrieving ROC curve...')
        roc = self._buffer_read(
            request_id=request_id, client_id=client_id, dataframe_name='roc',
            verbose=verbose)
        if 'TPR' in roc.keys():
            roc['TPR'] = roc['TPR'].astype(float)
        if 'FPR' in roc.keys():
            roc['FPR'] = roc['FPR'].astype(float)
            roc = roc.sort_values(by=['FPR'], ascending=True)

        return features, confusion_matrix, stats, roc



"""
Copyright (c) 2020-2022 Go2Market Insights, LLC d/b/a Analyzr.
All rights reserved.
https://analyzr.ai

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from .analyzer import Analyzer
# from .client_base import BaseClient



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name='analyzr-sdk-python',
    version='1.2.51',
    description='Python SDK for Analyzr API',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/analyzr-ai/analyzr-sdk-python",
    author='Analyzr Team',
    author_email='support@analyzr.ai',
    license='Apache 2.0',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=['analyzrclient'],
    install_requires=[
    ]
)



import os, sys, time, json
import pandas as pd
from copy import deepcopy

from .runner_base import BaseRunner
from .constants import *
from .utils import *

DEBUG = True

class ClusterRunner(BaseRunner):
    """
    Run the clustering pipeline

    :param client: SAML SSO client object
    :type client: SamlSsoAuthClient, required
    :param base_url: Base URL for the Analyzr API tenant
    :type base_url: str, required
    """

    def __init__(self, client=None, base_url=None):
        """
        """
        super().__init__(client=client, base_url=base_url)
        self.__uri = '{}/analytics/'.format(self._base_url)
        return

    def check_status(
            self, request_id=None, client_id=None, verbose=False,
            data=None):
        """
        Check the status of a specific model run. Data is homomorphically
        encoded by default

        :param request_id: UUID for a specific model object
        :type request_id: str, required
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :type client_id: string, required
        :param verbose: Set to true for verbose output
        :type verbose: boolean, optional
        :param data: if data is not None, cluster IDs will be appended and stats
            compiled
        :type data: DataFrame, optional
        :return: JSON object with the following attributes:
                    `status` (can be Pending, Complete, or Failed),
                    `request_id` (UUID provided with initial request),
                    `data` (dataframe with clustering results, if applicable)
        """
        res1 = {}
        res1['request_id'] = request_id
        res2 = self._status(
            request_id=request_id, client_id=client_id,
            verbose=verbose)
        if res2!={} and 'status' in res2.keys():
            res1['status'] = res2['status']
            if res2['status']=='Complete':
                # Load encoding keys
                keys = self._keys_load(model_id=request_id, verbose=verbose)
                if keys is None:
                    print('ERROR! Keys not found. ')
                    return None
                df2 = self.__retrieve_train_results(
                    request_id=request_id, client_id=client_id,
                    categorical_vars=list(keys['xref'].keys()),
                    numerical_vars=list(keys['zref'].keys()),
                    idx_var=keys['idx_var'], xref=keys['xref'],
                    zref=keys['zref'], rref=keys['rref'],
                    fref=keys['fref'], verbose=verbose)
                res1['data'] = df2
                if data is not None:
                    if DEBUG: print('Compiling stats...')
                    res3 = self.__post_process_results(
                        data, df2, keys['idx_var'],
                        list(keys['xref'].keys()))
                    res3['status'] = res1['status']
                    res3['request_id'] = request_id
                    res1 = res3
            if res2['status'] in ['Complete', 'Failed']:
                self._buffer_clear(
                    request_id=request_id, client_id=client_id,
                    verbose=verbose)

        return res1

    def __post_process_results(self, df, pc_id, idx_var, categorical_vars):
        """
        :param df:
        :param pc_id:
        :param idx_var:
        :param categorical_vars:
        :return res:
        """
        res = {}
        df3 = merge_cluster_ids(df, pc_id, idx_var)
        res['data'] = df3
        res['stats'] = compute_cluster_stats(df3, categorical_vars)
        res['distances'] = compute_cluster_distances(df3, categorical_vars)
        return res

    def run(
            self, df, client_id=None, idx_var=None, categorical_vars=[],
            numerical_vars=[], algorithm='pca-kmeans', n_components=5,
            buffer_batch_size=1000, cluster_batch_size=None,
            verbose=False, poll=True, compressed=False, staging=True):
        """
        Run clustering algorithm on user-provided dataset

        :param df: dataframe containing dataset to be clustered. The data is
            homomorphically encrypted by the client prior to being transferred
            to the API buffer
        :type df: DataFrame, required
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :type client_id: string, required
        :param idx_var: name of index field identifying unique record IDs in
            `df` for audit purposes
        :type idx_var: string, required
        :param categorical_vars: array of field names identifying categorical
            fields in the dataframe `df`
        :type categorical_vars: string[], required
        :param numerical_vars: array of field names identifying categorical
            fields in the dataframe `df`
        :type numerical_vars: string[], required
        :param algorithm: can be any of the following: `pca-kmeans`,
            `incremental-pca-kmeans`, `pca-kmeans-simple`, `kmeans`,
            `minibatch-kmeans`, `gaussian-mixture`, `birch`, `dbscan`, `optics`,
            `mean-shift`, `spectral-clustering`, `hierarchical-agglomerative`.
            Algorithms are sourced from Scikit-Learn unless otherwise indicated.
        :type algorithm: string, required
        :param n_components: number of clustering components
        :type n_components: int, optional
        :param buffer_batch_size: batch size for the purpose of uploading data
            from the client to the server's buffer :param buffer_batch_size:
            batch size for the purpose of uploading data from the client to the
            server's buffer
        :type buffer_batch_size: int, optional
        :param cluster_batch_size: batch size for the purpose of clustering the
            data provided in the dataframe `df`
        :type cluster_batch_size: int, optional
        :param verbose: Set to true for verbose output
        :type verbose: boolean, optional
        :param poll: keep polling API while the job is being run (default is
            `True`)
        :type poll: boolean, optional
        :param compressed: perform additional compression when uploading data
            to buffer
        :type compressed: boolean, optional
        :param staging: when set to True the API will use temporay secure cloud
            storage to buffer the data rather than a relational database
            (default is `True`)
        :type staging: boolean, optional
        :return: JSON object with the following attributes:
                    `model_id` (UUID provided with initial request),
                    `request_id`: same as `model_id` (provided for backward compatibility),
                    `data`: original dataset with cluster IDs appended
                    `distances`: distance matrix showing inter-cluster distances
                    (centroid to centroid)
                    `stats`: count, frequency, and attribute averages by cluster
                    ID
        """
        request_id = self._get_request_id()
        return self.__train(
            df, categorical_vars=categorical_vars, numerical_vars=numerical_vars,
            idx_var=idx_var, buffer_batch_size=buffer_batch_size,
            cluster_batch_size=cluster_batch_size, algorithm=algorithm,
            n_components=n_components, request_id=request_id, client_id=client_id,
            verbose=verbose, compressed=compressed, poll=poll, staging=staging,
        )

    def __train(
            self, df, categorical_vars=[], numerical_vars=[], bool_vars=[],
            idx_var=None, verbose=False, buffer_batch_size=1000,
            algorithm='pca-kmeans', n_components=5, cluster_batch_size=None,
            request_id=None, client_id=None, timeout=600, step=2,
            compressed=False, poll=True, staging=False):
        """
        """
        # Encode data and save it to buffer
        df3 = None
        if verbose: print('Request ID: {}'.format(request_id))
        data, xref, zref, rref, fref, fref_exp, bref = self._encode(
            df, categorical_vars=categorical_vars, numerical_vars=numerical_vars,
            bool_vars=bool_vars, record_id_var=idx_var, verbose=verbose)

        # Save encoding keys locally
        self._keys_save(
            model_id=request_id,
            keys={
                'xref': xref,
                'zref': zref,
                'rref': rref,
                'fref': fref,
                'fref_exp': fref_exp,
                'bref': bref,
                'idx_var': idx_var
            },
            verbose=verbose)

        # Save encoded data to buffer
        res = self._buffer_save(
            data, client_id=client_id, request_id=request_id, verbose=verbose,
            batch_size=buffer_batch_size, compressed=compressed, staging=staging)

        # Identify clusters and retrieve results
        if res['batches_saved']==res['total_batches']:
            self.__cluster_train(
                request_id=res['request_id'],
                client_id=client_id,
                idx_field=fref['forward'][idx_var],
                categorical_fields=[ fref['forward'][var] for var in categorical_vars ],
                algorithm=algorithm,
                n_components=n_components,
                batch_size=cluster_batch_size,
                verbose=verbose,
                staging=staging,
            )
            if poll:
                res2 = self._poll(
                    payload={
                        'request_id': res['request_id'],
                        'client_id': client_id,
                        'command': 'task-status'
                    },
                    timeout=timeout,
                    step=step,
                    verbose=verbose)
                if res2['response']['status'] in ['Complete']:
                    df2 = self.__retrieve_train_results(
                        request_id=res['request_id'], client_id=client_id,
                        categorical_vars=categorical_vars,
                        numerical_vars=numerical_vars, idx_var=idx_var,
                        xref=xref, zref=zref, rref=rref, fref=fref,
                        verbose=verbose)
                else:
                    print('WARNING! Training request came back with status: {}'.format(res2['response']['status']))
        else:
            print('ERROR! Buffer save failed: {}'.format(res))

        # Clear buffer
        if poll: res3 = self._buffer_clear(
            request_id=res['request_id'], client_id=client_id,
            verbose=verbose)

        #  Compile results
        if verbose and not poll: print('Clustering job started with polling disabled. You will need to request results for this request ID.')
        res5 = self.__post_process_results(
            df, df2, idx_var, categorical_vars) if poll else {}
        res5['request_id'] = request_id # append request ID for future reference
        res5['model_id'] = request_id # append request ID for future reference
        return res5

    def __cluster_train(
            self, request_id=None, client_id=None, idx_field=None,
            categorical_fields=[], algorithm='pca-kmeans', n_components=5,
            batch_size=None, verbose=False, staging=False):
        """
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :param idx_field: name of index field identifying unique record IDs for
            audit purposes
        :param categorical_fields:
        :param algorithm:
        :param n_components:
        :param batch_size:
        :param verbose: Set to true for verbose output
        :param staging:
        :return res:
        """
        if verbose: print('Clustering data in buffer...')
        if algorithm=='dask-pca-kmeans' and staging is False:
            print('WARNING! Found staging set to False with dask-pca-kmeans request. Setting staging to True.')
            staging = True
        # uri = '{}/analytics/'.format(self._base_url)
        res = self._client._post(self.__uri, {
            'command': 'cluster-lazy' if algorithm=='dask-pca-kmeans' else 'cluster-train',
            'request_id': request_id,
            'client_id': client_id,
            'algorithm': algorithm,
            'n_components': n_components,
            'batch_size': batch_size,
            'idx_field': idx_field,
            'categorical_fields': categorical_fields,
            'staging': staging,
        })
        return res

    def __retrieve_train_results(
            self, request_id=None, client_id=None, categorical_vars=[],
            numerical_vars=[], idx_var=None, xref=None, zref=None, rref=None,
            fref=None, verbose=False):
        """
        :return df2:
        """
        df2 = None
        res = self._buffer_read(
            request_id=request_id, client_id=client_id, dataframe_name='res',
            verbose=verbose, staging=True)
        if not res.empty:
            df2 = self._decode(
                res, categorical_vars=categorical_vars,
                numerical_vars=numerical_vars, record_id_var=idx_var,xref=xref,
                zref=zref, rref=rref, fref=fref, verbose=verbose)

        return df2

    def predict(
            self, df, model_id=None, client_id=None, idx_var=None,
            categorical_vars=[], numerical_vars=[], buffer_batch_size=1000,
            cluster_batch_size=None, timeout=600, verbose=False,
            compressed=False, staging=True):
        """
        Assign cluster IDs to dataset using a pre-trained model

        :param df: dataframe containing dataset to be clustered. The data is
            homomorphically encrypted by the client prior to being transferred
            to the API buffer
        :type df: DataFrame, required
        :param model_id: UUID for a specific model object
        :type model_id: str, required
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :type client_id: string, required
        :param idx_var: name of index field identifying unique record IDs in
            `df` for audit purposes
        :type idx_var: string, required
        :param categorical_vars: array of field names identifying categorical
            fields in the dataframe `df`
        :type categorical_vars: string[], required
        :param numerical_vars: array of field names identifying categorical
            fields in the dataframe `df`
        :type numerical_vars: string[], required
        :param buffer_batch_size: batch size for the purpose of uploading data
            from the client to the server's buffer :param buffer_batch_size:
            batch size for the purpose of uploading data from the client to the
            server's buffer
        :type buffer_batch_size: int, optional
        :param cluster_batch_size: batch size for the purpose of clustering the
            data provided in the dataframe `df`
        :type cluster_batch_size: int, optional
        :param timeout: client will keep polling API for a period of `timeout`
            seconds
        :type timeout: int, optional
        :param verbose: Set to true for verbose output
        :type verbose: boolean, optional
        :param compressed: perform additional compression when uploading data to
            buffer
        :type compressed: boolean, optional
        :param staging: when set to True the API will use temporay secure cloud
            storage to buffer the data rather than a relational database
            (default is `True`)
        :type staging: boolean, optional
        :return: JSON object with the following attributes:
                    `model_id` (UUID provided with initial request),
                    `data2`: original dataset with cluster IDs appended
        """
        return self.__predict(
            df, categorical_vars=categorical_vars, numerical_vars=numerical_vars,
            idx_var=idx_var, buffer_batch_size=buffer_batch_size,
            cluster_batch_size=cluster_batch_size, model_id=model_id,
            client_id=client_id, timeout=timeout, verbose=verbose,
            compressed=compressed, staging=staging,
        )

    def __predict(
            self, df, categorical_vars=[], numerical_vars=[], bool_vars=[],
            idx_var=None, verbose=False, buffer_batch_size=1000,
            cluster_batch_size=None, model_id=None, client_id=None, timeout=600,
            step=2, compressed=False, staging=False, encoding=True):
        """
        """
        request_id = self._get_request_id()

        if encoding:
            keys = self._keys_load(model_id=model_id, verbose=verbose) # Load encoding keys
            if keys is None:
                print('ERROR! Keys not found. ')
                return None
            data, xref, zref, rref, fref, fref_exp, bref = self._encode(
                df, keys=keys, categorical_vars=categorical_vars,
                numerical_vars=numerical_vars, bool_vars=bool_vars,
                record_id_var=idx_var, verbose=verbose)
            if verbose: print('Total rows encoded: {:,}'.format(len(data)))
        else:
            data = df

        # Save data to buffer
        res = self._buffer_save(
            data,
            client_id=client_id,
            request_id=request_id,
            verbose=verbose,
            batch_size=buffer_batch_size,
            compressed=compressed,
            staging=staging
        )

        # Predict clusters and retrieve results
        if res['batches_saved']==res['total_batches']:
            self.__cluster_predict(
                request_id=res['request_id'],
                model_id=model_id,
                client_id=client_id,
                idx_field=fref['forward'][idx_var] if encoding else idx_var,
                categorical_fields=[ fref['forward'][var] for var in categorical_vars ] if encoding else categorical_vars,
                batch_size=cluster_batch_size,
                verbose=verbose,
                staging=staging,
            )
            self._poll(
                payload={
                    'request_id': res['request_id'],
                    'client_id': client_id,
                    'command': 'task-status'
                },
                timeout=timeout,
                step=step,
                verbose=verbose)
            data2 = self.__retrieve_predict_results(
                request_id=request_id,
                client_id=client_id,
                verbose=verbose)
        else:
            print('ERROR! Buffer save failed: {}'.format(res))

        # Clear buffer
        res4 = self._buffer_clear(
            request_id=res['request_id'], client_id=client_id,
            verbose=verbose)

        # Decode data
        if encoding:
            data2 = self._decode(
                data2,
                categorical_vars=categorical_vars,
                numerical_vars=numerical_vars,
                bool_vars=bool_vars,
                record_id_var=idx_var,
                xref=xref,
                zref=zref,
                rref=rref,
                fref=fref,
                bref=bref,
                verbose=verbose
            )
        #  Compile results
        res5 = {}
        res5['data2'] = data2
        res5['model_id'] = model_id
        return res5

    def __cluster_predict(
            self, request_id=None, model_id=None, client_id=None, idx_field=None,
            categorical_fields=[], batch_size=None, verbose=False, staging=False):
        """
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting
            purposes only
        :param model_id:
        :param idx_field: name of index field identifying unique record IDs for
            audit purposes
        :param categorical_fields:
        :param batch_size:
        :param verbose: Set to true for verbose output
        :param staging:
        :return res:
        """
        if verbose: print('Predicting cluster IDs using data in buffer...')
        res = self._client._post(self.__uri, {
            'command': 'cluster-predict',
            'model_id': model_id,
            'request_id': request_id,
            'client_id': client_id,
            'batch_size': batch_size,
            'idx_field': idx_field,
            'categorical_fields': categorical_fields,
            'staging': staging,
        })
        return res

    def __retrieve_predict_results(
            self, request_id=None, client_id=None,
            verbose=False):
        """
        :return data2:
        """
        data2 = self._buffer_read(
            request_id=request_id, client_id=client_id, dataframe_name='res',
            verbose=verbose, staging=True)
        return data2



import os, sys, time, json
import pandas as pd
from copy import deepcopy
import datetime

from .client_saml_sso import SamlSsoAuthClient
from .constants import *
from .utils import *
from .runner_cluster import ClusterRunner
from .runner_propensity import PropensityRunner
from .runner_regression import RegressionRunner
from .runner_task import TaskRunner


class Analyzer:
    """
    Parent class for Analyzr client. This is the class that should be instantiated
    by the client. For detailed methods, see the appropriate runner class.

    :param host: the FQDN for your API tenant
    :type host: str, required
    :verbose: Set to true for verbose output
    :type verbose: boolean, optional
    """

    def __init__(self, host=None, verbose=False):
        """
        """
        # General config
        self.__client = SamlSsoAuthClient(host=host, verbose=verbose)
        self.__base_url = 'https://{}/api/v1'.format(host)
        self.__uri = '{}/analytics/'.format(self.__base_url)
        self.test = TaskRunner(client=self.__client, base_url=self.__base_url)

        # Initialize specific endpoints
        self.cluster = ClusterRunner(client=self.__client, base_url=self.__base_url)
        self.propensity = PropensityRunner(client=self.__client, base_url=self.__base_url)
        self.regression = RegressionRunner(client=self.__client, base_url=self.__base_url)
        return

    def version(self):
        """
        Provide API version info

        :return: JSON object with API version and other metadata
        """
        return self.__client._post(self.__uri, {'command': 'version'})

    def client_version(self):
        """
        Provide client version info

        :return: JSON object with API version and other metadata
        """
        return ({'version': f'{CLIENT_VERSION}', 'copyright': f'{datetime.date.today().year} (c) Go2Market Insights Inc. All rights reserved. Patent pending. '})

    def login(self, verbose=False):
        """
        Log in to Analyzr API

        :param verbose: Set to true for verbose output Set to True for verbose screen output
        :type verbose: boolean, optional
        :rtype: None
        """
        status_code = self.__client._login(verbose=verbose)
        if status_code==200:
            print('Login successful')
        else:
            print('Could not log in (status code: {})'.format(status_code))

        return

    def logout(self, verbose=False):
        """
        Log out of Analyzr API

        :param verbose: Set to true for verbose output Set to True for verbose screen output
        :type verbose: boolean, optional
        :rtype: None
        """
        return self.__client._logout(verbose=verbose)



"""
Copyright (c) 2020-2021 Go2Market Insights, LLC
All rights reserved.
https://g2m.ai

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os, sys, time, json
import pandas as pd
from copy import deepcopy

from .runner_base import BaseRunner
from .constants import *
from .utils import *

class TaskRunner(BaseRunner):
    """
    Runs pipeline for generic tasks (asynchronous testing)

    """
    def __init__(self, client=None, base_url=None):
        """
        """
        super().__init__(client=client, base_url=base_url)
        self.__uri = '{}/analytics/'.format(self._base_url)
        return

    def run(self, type='simple', verbose=False, compressed=False):
        """
        Returns test task output

        :param type: can be 'simple' (default) or 'storage'
        :param verbose: Set to true for verbose output
        :param compressed:
        :return res:
        """
        request_id = self._get_request_id()
        return self.__run(df=get_test_data(), request_id=request_id, client_id='test', type=type, verbose=verbose, compressed=compressed)

    def __run(self, df=None, request_id=None, client_id=None, type='simple', verbose=False, compressed=False):
        """
        :param df:
        :param request_id:
        :param client_id: Short name for account being used. Used for reporting purposes only
        :param type: can be 'simple' (default), 'storage', or 'bz2'
        :param verbose: Set to true for verbose output
        :param comnpressed:
        :return res:
        """
        client_id = 'test'
        payload = {'request_id': request_id, 'client_id': client_id, 'type': type, 'command': 'test-task', 'data': df.to_dict()}
        self._client._post(self.__uri, payload, compressed=compressed)
        self._poll(payload={'request_id': request_id, 'client_id': client_id, 'command': 'task-status'}, timeout=60, step=1, verbose=verbose)
        res = self._buffer_read(request_id=request_id, client_id=client_id, verbose=verbose, raw=True)
        return res if res['status']!=200 else {
            'status': res['status'],
            'response': {
                'request_id': res['response']['request_id'],
                'client_id': res['response']['client_id'],
                'data': pd.DataFrame(res['response']['data']),
                'type': type,
            }
        }
